import PIL
import piexif
import pyheif
import PIL.Image
import PIL.ExifTags

from sortmedia.file import File
from datetime import datetime


class Photo(File):
    def creation_date(self):
        """ Returns the creation date based on the file's EXIF data. """
        super().creation_date()

        # If we do not have exif data, return `None`.
        if not self.exif:
            return None

        keys = ["DateCreated",
                "DateTimeCreated",
                "DateTimeOriginal",
                "DateTime",
                "DateTimeDigitized",
                "DigitalCreationDateTime"]

        created_str = None

        for k in keys:
            if k in self.exif:
                created_str = self.exif[k]
                break

        if isinstance(created_str, bytes):
            created_str = created_str.decode('utf-8')

        parsed = datetime.strptime(created_str, "%Y:%m:%d %H:%M:%S")

        return {
            'year': str(parsed.year),
            'month': str(parsed.strftime("%B")),
            'day': str(parsed.day)
        }

    def get_exif(self):
        """ Returns the EXIF data for a photo. Includes support for HEIC file
            format. """
        exif = {}

        # iOS now stores images in `heic`. PIL doesn't support this format.
        # This allows us to use the bindings to libheif to grab the EXIF data.
        if self.mime == "image/heic":
            heif = pyheif.read_heif(self.fp)
            for metadata in heif.metadata or []:
                if metadata['type'] == 'Exif':
                    data = metadata.get('data', {})
                    exif = piexif.load(data).get('Exif').items()

        # Otherwise, use PIL to process the image.
        else:
            img = PIL.Image.open(self.fp)
            exif = img._getexif().items()

        self.fp.seek(0)
        return {PIL.ExifTags.TAGS[k]: v for k, v in exif if k in
                PIL.ExifTags.TAGS}



import argparse
from sortmedia.sort import SortMedia

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c',
                        '--copy',
                        action='store_true',
                        help='Copy instead of move files',
                        required=False,
                        default=False)
    parser.add_argument('-d',
                        '--dry',
                        action='store_true',
                        help='Dry run - do not actually process files',
                        required=False,
                        default=False)
    parser.add_argument('src',
                        nargs='+',
                        help='The media source directory to process')
    parser.add_argument('dst',
                        nargs='+',
                        help='The destination directory to move or copy files')
    parser.add_argument('-n',
                        '--noprocess',
                        help='List of directories to move but not recurse into ' +
                             'for processing',
                        type=str)
    parser.add_argument('-e',
                        '--excludes',
                        help='List of directories to skip completely',
                        type=str)
    parser.add_argument('-w',
                        '--watch',
                        action='store_true',
                        help='Watch the source directory for new files.',
                        required=False,
                        default=False)
                        

    args = parser.parse_args()

    src = args.src[0]
    dst = args.dst[0]

    if args.noprocess:
        args.noprocess = [x for x in args.noprocess.split(',')]

    if args.excludes:
        args.excludes = [x for x in args.excludes.split(',')]

    ms = SortMedia(dry=args.dry, 
                   copy=args.copy,
                   noprocess=args.noprocess,
                   excludes=args.excludes,
                   watch=args.watch)
    ms.process_files(src, dst)



import os
import shutil
import hashlib
import logging

from sortmedia.util import Safety
from abc import ABC

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class File(ABC):
    """ Represents a media file on disk. Tracks the MIME type, EXIF data and
        an open file handle. Provides several functions for interacting with
        the EXIF data and moving/copying the files. """

    def __init__(self, path, mime, exif=None):
        """ Create a new `File` object for the file data located at `path` and
            which is represented by the open file pointer `fp`.

            @param  path    The path to the target file.
            @param  mime    The string MIME type of the file.
            @param  exif    A dictionary of EXIF value. """

        if not os.path.exists(path):
            raise RuntimeError(f"The path `{path}` does not exist.")

        self.path = path
        self.fp = open(path, 'rb')
        self.hash = self.get_hash()
        self.exif = None
        self.mime = mime

        super().__init__()

    def __del__(self):
        """ Closes the open file pointer on object destruct. """
        self.fp.close()

    def get_hash(self):
        """ Returns the hash of the file pointer. """
        m = hashlib.sha256()
        m.update(self.fp.read())
        h = m.digest()
        self.fp.seek(0)
        return h

    def creation_date(self):
        """ This method should be implemented by the inheriting class. The
            child class should call this function to ensure that the EXIF data
            is populated. """
        if not self.exif:
            self.exif = self.get_exif()

    def __path_from_date(self, root, name):
        """ Returns a path based on the given creation date, in the format of:
            `<root>/<year>/<month>/<name>`.

            @param  root    The root of our absolute path.
            @param  name    The filename of our path.

            @returns    An absolute path. """
        creation_date = self.creation_date()

        if not creation_date:
            return os.path.join(root, 'unknown', name)

        return os.path.join(root,
                            creation_date["year"],
                            creation_date["month"],
                            creation_date["day"],
                            name)

    def hashes_match(self, path):
        """ Returns a boolean indicating if the hash of `path` matches the
            object hash.

            @param  path    The path to compare hashes with.

            @returns    `True` if the path's hash matches the objects hash,
                        `False` otherwise. """
        with open(path, "rb") as f:
            m = hashlib.sha256()
            m.update(f.read())
            return self.hash == m.digest()

    def write_check(self, path):
        """ Runs some preliminary checks to ensure that we do not accidentally
            overwrite an existing file. If a file exists at `path` we check
            its hash. If the hashes match, we return `IDENTICAL`. If they do
            not match, we return `UNSAFE` - we do not want to write the file
            to this path. If no file exists at `path` we return `SAFE`.

            @param  path    The path to check.

            @returns    `Safety` """
        if os.path.exists(path):
            if not self.hashes_match(path):
                return Safety.UNSAFE
            else:
                return Safety.IDENTICAL
        return Safety.SAFE

    def target_path(self, root):
        """ Creates a "safe" target path by performing some basic checks. If
            a file exists at the target path with a non-identical hash we
            create a new file with a `_N` prefix (where `N` is an integer
            value).

            @param  root    The root path to set the target path in.

            @returns    An absolute path to write a file to. """
        name, ext = os.path.splitext(os.path.basename(self.fp.name))
        mod = ''
        i = 1
        path = self.__path_from_date(root, f"{name}{mod}{ext}")

        check = self.write_check(path)

        while check == Safety.UNSAFE:
            mod = f'_{i}'

            path = self.__path_from_date(root, f"{name}{mod}{ext}")
            check = self.write_check(path)
            i = i + 1

        if check == Safety.SAFE:
            return path

        return None

    def get_exif(self):
        """ Returns the EXIF data for the file. This should be implemented
            in the inheriting class. """
        raise NotImplementedError()

    def make_nested_dirs(self, target):
        """ Create nested directories if they do not exist. """
        dirs = os.path.dirname(target)

        if not os.path.exists(dirs):
            os.makedirs(dirs)

    def move(self, root):
        """ Safely move this file to the target path. """
        target = self.target_path(root)

        if target:
            self.make_nested_dirs(target)
            shutil.move(self.fp.name, target)
            return True

        return False

    def copy(self, root):
        """ Safely copy this file to the target path. """
        target = self.target_path(root)

        if target:
            self.make_nested_dirs(target)
            shutil.copyfile(self.fp.name, target)
            return True

        return False



import magic
import filetype
from enum import Enum

class Safety(Enum):
    SAFE = 1
    UNSAFE = 2
    IDENTICAL = 3


def get_mimetype(path):
    """ Get the mimetype of the given file pointer.
        @param  path    Path to a file to get the MIME for.
        @return     A string representing the mimetype or `None` if no MIME
                    coul be discerned. """
    guess = filetype.guess(path)

    if guess:
        return guess.mime

    # If we were unable to guess the MIME, attempt to use filemagic.
    with magic.Magic() as m:
        return m.id_filename(path)

def is_video(mime):
    """ Returns `True` if the specified file pointer represents a video.
        `False` otherwise. """
    return mime.startswith("video/")


def is_photo(mime):
    """ Returns `True` if the specified file pointer represents a photo.
        `False` otherwise. """
    return mime.startswith("image/")



import os
import shutil
import logging

from pathlib import Path
from sortmedia.photo import Photo
from sortmedia.video import Video
from sortmedia.util import is_video, is_photo, get_mimetype

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class SortMedia:
    """ Handles processing the specified source path. """

    def __init__(self, 
                 dry=False, 
                 copy=False, 
                 noprocess=[], 
                 excludes=[], 
                 watch=False):
        """ Initialize a new media sorter.
        
            @param  dry         Print out files that would be processed without
                                actually processing anything.
            @param  copy        Copy instead of move files.
            @param  noprocess   List of directories to skip processing, but
                                still move.
            @param  excludes    List of directories to not process. 
            @param  watch       Watch the source directory for changes. """
        logger.info(f"Dry Run: {'Yes' if dry else 'No'}")
        logger.info(f"Mode: {'Copy' if copy else 'Move'}")
        logger.info(f"Watch: {'Yes' if watch else 'No'}")
        self.copy = copy
        self.dry_run = dry
        self.noprocess = noprocess
        self.excludes = excludes
        self.watch = watch

    def is_exclude_dir(self, path):
        """ Test if the filepath is located in one of the specified exclude 
            directories. 
            
            @param  path    The path to test.
            
            @returns    `True` if the path should not be processed. `False`
                        otherwise. """
        if not self.excludes:
            return

        for parent in self.excludes:
            x = os.path.commonpath([parent, path]) 
            if os.path.normpath(x) == os.path.normpath(parent):
                return True 

        return False

    def handle_no_process_dirs(self, dst):
        """ Directories marked as "no process" should be moved/copied into the
            target directory without processing the files. """
        if not self.noprocess:
            return

        for p in self.noprocess:
            dst_path = os.path.join(dst, p)

            if self.copy:
                shutil.copytree(p, dst_path) 
            else:
                shutil.move(p, dst_path)

    def process_files(self, src, dst):
        """ Kick off processing.
        
            @param  src     The source directory to process.
            @param  dst     The destination to move files to. """
        if not os.path.isdir(src):
            logger.error(f"The specified source director `{src}` doest not " +
                         "exist.")
            return

        if self.watch:
            logger.info(f"After initial processing, monitoring will begin " +
                        "on `{src}`. Ctrl+C to stop monitoring.")

        total_photos = 0
        total_videos = 0
        total_duplicates = 0
        skipped = 0

        self.handle_no_process_dirs(dst) 

        for f in Path(src).rglob("*"):
            if not os.path.isfile(f):
                continue

            filepath = f.as_posix()

            if self.is_exclude_dir(filepath):
                skipped = skipped + 1
                logger.info(f"{filepath} -- NO PROCESS")
                continue

            mime = get_mimetype(filepath)
            obj = None

            if is_video(mime):
                obj = Video(filepath, mime)
                total_videos = total_videos + 1
            elif is_photo(mime):
                obj = Photo(filepath, mime)
                total_photos = total_photos + 1

            if obj:
                if self.dry_run:
                    logger.info(filepath)
                else:
                    took_action = False
                    if not self.copy:
                        took_action = obj.move(dst)
                    else:
                        took_action = obj.copy(dst)

                    if not took_action:
                        total_duplicates = total_duplicates + 1

                    detail = ' -- DUPLICATE' if not took_action else ''
                    logger.info(f"{filepath}{detail}")

        
        logger.info(f"Processed {total_photos+total_videos+skipped} files")
        logger.info(f"\tTotal videos...: {total_videos}")
        logger.info(f"\tTotal photos...: {total_photos}")
        logger.info(f"\tDuplicates.....: {total_duplicates}")
        logger.info(f"\tExcluded.......: {skipped}")

        if self.watch:
            logger.info("Going into monitor state. Any new files added to " +
                        "source directory will be automatically processed.")
            self.monitor(src, dst)

    def monitor(self, src, dst):
        """ """
        pass



#!/usr/bin/env python

from os import path
from setuptools import setup

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='sortmedia',
      version='1.2.1',
      description='Dedupe and sort photo and video files based on the date ' +
                  'they were created.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Louis Lang',
      author_email='louis@lang.sh',
      url='https://github.com/LouisLang/sortmedia',
      keywords='file-sorting media-manager photos videos',
      packages=['sortmedia'],
      entry_points={
        'console_scripts': [
            'sortmedia=sortmedia.cli:main',
        ],
      },
      install_requires=requirements,
     )



import json

from datetime import datetime
from sortmedia.file import File
from subprocess import check_output


class Video(File):
    def creation_date(self):
        """ Returns the creation date based on the file's EXIF data. """
        super().creation_date()

        created_str = self.exif.get('creation_time')

        if not created_str:
            return None

        parsed = datetime.strptime(created_str, "%Y-%m-%dT%H:%M:%S.000000Z")
        return {
            'year': str(parsed.year),
            'month': str(parsed.strftime("%B")),
            'day': str(parsed.day)
        }

    def get_exif(self):
        """ Returns the EXIF data for a photo. Includes support for HEIC file
            format. """
        cmd = ['ffprobe', '-v', 'quiet', self.fp.name, '-print_format', 'json',
               '-show_entries',
               'stream=index,codec_type:stream_tags=creation_time:' +
               'format_tags=creation_time']
        out = check_output(cmd)
        j = json.loads(out)
        streams = j.get("streams", {})

        if len(streams) == 0:
            return None

        return streams[0].get("tags", None)
